# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feature_store.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import reflection as _reflection
from google.protobuf import message as _message
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x66\x65\x61ture_store.proto\x12\x1b\x66\x65\x61ture_store.common.protos\"$\n\x14SourceDataColumnInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x84\x01\n\x11\x46\x65\x61tureColumnInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\t\x12\x12\n\nlookup_key\x18\x03 \x03(\t\x12\x13\n\x0boutput_name\x18\x04 \x01(\t\x12\x1c\n\x14timestamp_lookup_key\x18\x05 \x03(\t\"~\n\x12OnDemandColumnInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\x12\x41\n\x0einput_bindings\x18\x02 \x03(\x0b\x32).feature_store.common.protos.InputBinding\x12\x13\n\x0boutput_name\x18\x03 \x01(\t\"3\n\x0cInputBinding\x12\x11\n\tparameter\x18\x01 \x01(\t\x12\x10\n\x08\x62ound_to\x18\x02 \x01(\t\"\xfc\x02\n\nColumnInfo\x12T\n\x17source_data_column_info\x18\x01 \x01(\x0b\x32\x31.feature_store.common.protos.SourceDataColumnInfoH\x00\x12M\n\x13\x66\x65\x61ture_column_info\x18\x02 \x01(\x0b\x32..feature_store.common.protos.FeatureColumnInfoH\x00\x12P\n\x15on_demand_column_info\x18\x03 \x01(\x0b\x32/.feature_store.common.protos.OnDemandColumnInfoH\x00\x12\x14\n\x07include\x18\x04 \x01(\x08H\x01\x88\x01\x01\x12\x11\n\tdata_type\x18\x05 \x01(\t\x12!\n\x14topological_ordering\x18\x06 \x01(\x05H\x02\x88\x01\x01\x42\x06\n\x04infoB\n\n\x08_includeB\x17\n\x15_topological_ordering\"Q\n\x10\x46\x65\x61tureTableInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x10\n\x08table_id\x18\x02 \x01(\t\x12\x17\n\x0flookback_window\x18\x03 \x01(\x01\"3\n\x15\x46unctionParameterInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\"\xa2\x01\n\x0c\x46unctionInfo\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12H\n\x0cinput_params\x18\x02 \x03(\x0b\x32\x32.feature_store.common.protos.FunctionParameterInfo\x12\x1a\n\x12routine_definition\x18\x03 \x01(\t\x12\x19\n\x11\x65xternal_language\x18\x04 \x01(\t\"\xb1\x02\n\x0b\x46\x65\x61tureSpec\x12>\n\rinput_columns\x18\x01 \x03(\x0b\x32\'.feature_store.common.protos.ColumnInfo\x12\x43\n\x0cinput_tables\x18\x02 \x03(\x0b\x32-.feature_store.common.protos.FeatureTableInfo\x12\x42\n\x0finput_functions\x18\x03 \x03(\x0b\x32).feature_store.common.protos.FunctionInfo\x12\x14\n\x0cworkspace_id\x18\x04 \x01(\x03\x12$\n\x1c\x66\x65\x61ture_store_client_version\x18\x05 \x01(\t\x12\x1d\n\x15serialization_version\x18\x06 \x01(\x05\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_store_pb2', globals())
_FEATURESPEC = DESCRIPTOR.message_types_by_name['FeatureSpec']
FeatureSpec = _reflection.GeneratedProtocolMessageType('FeatureSpec', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESPEC,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureSpec)
})
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SOURCEDATACOLUMNINFO._serialized_start=52
  _SOURCEDATACOLUMNINFO._serialized_end=88
  _FEATURECOLUMNINFO._serialized_start=91
  _FEATURECOLUMNINFO._serialized_end=223
  _ONDEMANDCOLUMNINFO._serialized_start=225
  _ONDEMANDCOLUMNINFO._serialized_end=351
  _INPUTBINDING._serialized_start=353
  _INPUTBINDING._serialized_end=404
  _COLUMNINFO._serialized_start=407
  _COLUMNINFO._serialized_end=787
  _FEATURETABLEINFO._serialized_start=789
  _FEATURETABLEINFO._serialized_end=870
  _FUNCTIONPARAMETERINFO._serialized_start=872
  _FUNCTIONPARAMETERINFO._serialized_end=923
  _FUNCTIONINFO._serialized_start=926
  _FUNCTIONINFO._serialized_end=1088
  _FEATURESPEC._serialized_start=1091
  _FEATURESPEC._serialized_end=1396
# @@protoc_insertion_point(module_scope)
